"""System Bridge: Modules Update"""
import asyncio
from collections.abc import Callable
from collections.abc import Awaitable

from systembridgeshared.base import Base
from systembridgeshared.database import Database

from systembridgebackend.modules.battery.update import BatteryUpdate
from systembridgebackend.modules.bridge.update import BridgeUpdate
from systembridgebackend.modules.cpu.update import CPUUpdate
from systembridgebackend.modules.disk.update import DiskUpdate
from systembridgebackend.modules.display.update import DisplayUpdate
from systembridgebackend.modules.gpu.update import GPUUpdate
from systembridgebackend.modules.memory.update import MemoryUpdate
from systembridgebackend.modules.network.update import NetworkUpdate
from systembridgebackend.modules.sensors.update import SensorsUpdate
from systembridgebackend.modules.system.update import SystemUpdate


class Update(Base):
    """Modules Update"""

    def __init__(
        self,
        database: Database,
    ) -> None:
        """Initialize"""
        super().__init__()
        self._database = database  # pylint: disable=duplicate-code

        self._classes = [
            {"name": "battery", "cls": BatteryUpdate(self._database)},
            {"name": "disk", "cls": DiskUpdate(self._database)},
            {"name": "system", "cls": SystemUpdate(self._database)},
        ]
        self._classes_frequent = [
            {"name": "cpu", "cls": CPUUpdate(self._database)},
            {"name": "display", "cls": DisplayUpdate(self._database)},
            {"name": "gpu", "cls": GPUUpdate(self._database)},
            {"name": "memory", "cls": MemoryUpdate(self._database)},
            {"name": "network", "cls": NetworkUpdate(self._database)},
        ]
        BridgeUpdate(self._database)

    async def _update(
        self,
        class_obj: dict,
        updated_callback: Callable[[str], Awaitable[None]],
    ) -> None:
        """Update"""
        await class_obj["cls"].update_all_data()
        await updated_callback(class_obj["name"])

    async def update_data(
        self,
        updated_callback: Callable[[str], Awaitable[None]],
    ) -> None:
        """Update Data"""
        self._logger.info("Update data")
        if not self._database.connected:
            self._database.connect()

        tasks = [self._update(cls, updated_callback) for cls in self._classes]
        await asyncio.gather(*tasks)

        self._logger.info("Finished updating data")

    async def update_frequent_data(
        self,
        updated_callback: Callable[[str], Awaitable[None]],
    ) -> None:
        """Update Data"""
        self._logger.info("Update frequent data")
        if not self._database.connected:
            self._database.connect()

        sensors_update = SensorsUpdate(self._database)
        await sensors_update.update_all_data()
        await updated_callback("sensors")

        tasks = [self._update(cls, updated_callback) for cls in self._classes_frequent]
        await asyncio.gather(*tasks)

        self._logger.info("Finished updating frequent data")



"""System Bridge: Update Disk"""
import asyncio

from systembridgeshared.database import Database

from systembridgebackend.modules.base import ModuleUpdateBase
from systembridgebackend.modules.disk import Disk


class DiskUpdate(ModuleUpdateBase):
    """Disk Update"""

    def __init__(
        self,
        database: Database,
    ) -> None:
        """Initialize"""
        super().__init__(database, "disk")
        self._disk = Disk()

    async def update_io_counters(self) -> None:
        """Update IO counters"""
        for key, value in self._disk.io_counters()._asdict().items():
            self._database.write("disk", f"io_counters_{key}", value)

    async def update_io_counters_per_disk(self) -> None:
        """Update IO counters per disk"""
        for key, value in self._disk.io_counters_per_disk().items():  # type: ignore
            for subkey, subvalue in value._asdict().items():
                self._database.write(
                    "disk", f"io_counters_per_disk_{key}_{subkey}", subvalue
                )

    async def update_partitions(self) -> None:
        """Update partitions"""
        device_list = []
        partition_list = []
        for partition in self._disk.partitions():
            if partition.device not in device_list:
                device_list.append(partition.device)
            partition_list.append(partition.mountpoint)
            for key, value in partition._asdict().items():
                self._database.write(
                    "disk",
                    f"partitions_{partition.mountpoint}_{key}",
                    value,
                )
        self._database.write("disk", "devices", device_list)
        self._database.write("disk", "partitions", partition_list)

    async def update_usage(self) -> None:
        """Update usage"""
        for partition in self._disk.partitions():
            if data := self._disk.usage(partition.mountpoint):
                for key, value in data._asdict().items():
                    self._database.write(
                        "disk", f"usage_{partition.mountpoint}_{key}", value
                    )

    async def update_all_data(self) -> None:
        """Update data"""
        await asyncio.gather(
            *[
                self.update_io_counters(),
                self.update_io_counters_per_disk(),
                self.update_partitions(),
                self.update_usage(),
            ]
        )



"""System Bridge: Sensors"""
from __future__ import annotations

import json
import subprocess
import sys

import psutil
from systembridgeshared.base import Base


class Sensors(Base):
    """Sensors"""

    def fans(self) -> dict | None:
        """Get fans"""
        if not hasattr(psutil, "sensors_fans"):
            return None
        return psutil.sensors_fans()  # type: ignore

    def temperatures(self) -> dict | None:
        """Get temperatures"""
        if not hasattr(psutil, "sensors_temperatures"):
            return None
        return psutil.sensors_temperatures()  # type: ignore

    def windows_sensors(self) -> dict | None:
        """Get windows sensors"""
        if sys.platform != "win32":
            return None

        try:
            # Import here to not raise error when importing file on linux
            # pylint: disable=import-error, import-outside-toplevel
            from systembridgewindowssensors import get_windowssensors_path
        except (ImportError, ModuleNotFoundError) as error:
            self._logger.error("Windows sensors not found: %s", error)
            return None

        path = get_windowssensors_path()

        self._logger.debug("Windows sensors path: %s", path)
        try:
            with subprocess.Popen(
                [path],
                stdout=subprocess.PIPE,
            ) as pipe:
                result = pipe.communicate()[0].decode()
            self._logger.debug("Windows sensors result: %s", result)
        except Exception as error:  # pylint: disable=broad-except
            self._logger.error("Windows sensors error: %s", error)
            return None

        try:
            return json.loads(result)
        except json.decoder.JSONDecodeError as error:
            self._logger.error(error)
            return None



"""System Bridge: Autostart"""
import platform


def autostart_disable():
    """Disable autostart"""
    if "Windows" in platform.system():
        # pylint: disable=import-error, import-outside-toplevel
        from systembridgebackend.autostart.windows import autostart_windows_disable

        autostart_windows_disable()
    elif "Linux" in platform.system():
        # pylint: disable=import-error, import-outside-toplevel
        from systembridgebackend.autostart.linux import autostart_linux_disable

        autostart_linux_disable()


def autostart_enable():
    """Enable autostart"""
    if "Windows" in platform.system():
        # pylint: disable=import-error, import-outside-toplevel
        from systembridgebackend.autostart.windows import autostart_windows_enable

        autostart_windows_enable()
    elif "Linux" in platform.system():
        # pylint: disable=import-error, import-outside-toplevel
        from systembridgebackend.autostart.linux import autostart_linux_enable

        autostart_linux_enable()



"""System Bridge: Options"""
from collections import defaultdict

from sanic import Sanic, response
from sanic_routing.router import Route

from systembridgebackend.server.cors import _add_cors_headers


def _compile_routes_needing_options(routes: dict[str, Route]) -> dict[str, frozenset]:
    """Compile a list of routes that need OPTIONS."""
    needs_options = defaultdict(list)
    for route in routes.values():
        if "OPTIONS" not in route.methods:
            needs_options[route.uri].extend(route.methods)

    return {uri: frozenset(methods) for uri, methods in dict(needs_options).items()}


def _options_wrapper(handler, methods):
    """Wrap a handler with CORS headers."""

    def wrapped_handler(request, *args, **kwargs):
        """Wrapper for handler."""
        nonlocal methods
        return handler(request, methods)

    return wrapped_handler


async def options_handler(_, methods) -> response.HTTPResponse:
    """Handle OPTIONS requests."""
    resp = response.empty()
    _add_cors_headers(resp, methods)
    return resp


def setup_options(app: Sanic, _):
    """Setup OPTIONS handlers."""
    app.router.reset()
    needs_options = _compile_routes_needing_options(app.router.routes_all)  # type: ignore
    for uri, methods in needs_options.items():
        app.add_route(
            _options_wrapper(options_handler, methods),
            uri,
            methods=["OPTIONS"],
        )
    app.router.finalize()



"""System Bridge: Update Memory"""
import asyncio
from systembridgeshared.database import Database

from systembridgebackend.modules.base import ModuleUpdateBase
from systembridgebackend.modules.memory import Memory


class MemoryUpdate(ModuleUpdateBase):
    """Memory Update"""

    def __init__(
        self,
        database: Database,
    ) -> None:
        """Initialize"""
        super().__init__(database, "memory")
        self._memory = Memory()

    async def update_swap(self) -> None:
        """Update Swap Memory"""
        for key, value in self._memory.swap()._asdict().items():
            self._database.write("memory", f"swap_{key}", value)

    async def update_virtual(self) -> None:
        """Update Virtual Memory"""
        for key, value in self._memory.virtual()._asdict().items():
            self._database.write("memory", f"virtual_{key}", value)

    async def update_all_data(self) -> None:
        """Update data"""
        await asyncio.gather(
            *[
                self.update_swap(),
                self.update_virtual(),
            ]
        )



"""System Bridge: Server Handler - Power"""
import asyncio
import os
import sys

from sanic.request import Request
from sanic.response import HTTPResponse, json
from systembridgeshared.settings import Settings


def sleep() -> None:
    """Send the system to sleep."""
    if sys.platform == "linux":
        os.system("systemctl suspend")
    elif sys.platform == "win32":
        os.system("rundll32.exe powrprof.dll,SetSuspendState 0,1,0")


def hibernate() -> None:
    """Hibernate the system."""
    if sys.platform == "linux":
        os.system("systemctl hibernate")
    elif sys.platform == "win32":
        os.system("rundll32.exe powrprof.dll,SetSuspendState 1,1,0")


def restart() -> None:
    """Restart the system."""
    if sys.platform == "linux":
        os.system("systemctl restart")
    elif sys.platform == "win32":
        os.system("shutdown /r /t 0")


def shutdown() -> None:
    """Shutdown the system."""
    if sys.platform == "linux":
        os.system("systemctl poweroff")
    elif sys.platform == "win32":
        os.system("shutdown /s /t 0")


def lock() -> None:
    """Lock the system."""
    if sys.platform == "linux":
        os.system("xdg-screensaver lock")
    elif sys.platform == "win32":
        os.system("rundll32.exe user32.dll,LockWorkStation")


def logout() -> None:
    """Logout the user."""
    if sys.platform == "linux":
        os.system("pkexec --user $(whoami) gdm-force-logout")
    elif sys.platform == "win32":
        os.system("logoff")


async def handler_sleep(
    _: Request,
    settings: Settings,
) -> HTTPResponse:
    """Handle sleep requests."""
    asyncio.get_running_loop().call_later(2, sleep)
    return json(
        {
            "message": "Sleeping",
        }
    )


async def handler_hibernate(
    _: Request,
    settings: Settings,
) -> HTTPResponse:
    """Handle hibernate requests."""
    asyncio.get_running_loop().call_later(2, hibernate)
    return json(
        {
            "message": "Hibernating",
        }
    )


async def handler_restart(
    _: Request,
    settings: Settings,
) -> HTTPResponse:
    """Handle restart requests."""
    asyncio.get_running_loop().call_later(2, restart)
    return json(
        {
            "message": "Restarting",
        }
    )


async def handler_shutdown(
    _: Request,
    settings: Settings,
) -> HTTPResponse:
    """Handle shutdown requests."""
    asyncio.get_running_loop().call_later(2, shutdown)
    return json(
        {
            "message": "Shutting down",
        }
    )


async def handler_lock(
    _: Request,
    settings: Settings,
) -> HTTPResponse:
    """Handle lock requests."""
    asyncio.get_running_loop().call_later(2, lock)
    return json(
        {
            "message": "Locking",
        }
    )


async def handler_logout(
    _: Request,
    settings: Settings,
) -> HTTPResponse:
    """Handle logout requests."""
    asyncio.get_running_loop().call_later(2, logout)
    return json(
        {
            "message": "Logging out",
        }
    )



"""System Bridge: Bridge"""
from collections.abc import Callable

from systembridgeshared.base import Base
from zeroconf import InterfaceChoice, ServiceBrowser, Zeroconf


class Bridge(Base):
    """Bridge"""

    def __init__(
        self,
        callback: Callable,
    ):
        super().__init__()
        ServiceBrowser(
            Zeroconf(
                interfaces=InterfaceChoice.All,
                unicast=True,
            ),
            "_system-bridge._tcp.local.",
            handlers=[callback],
        )



"""System Bridge: Server Handler - Media"""
from __future__ import annotations

import asyncio
from collections.abc import Callable
from datetime import datetime
import io
import mimetypes
import os
import re
import tempfile
from urllib.parse import urlencode

import aiofiles
from aiohttp import ClientSession
from mutagen import File as MutagenFile
from plyer import storagepath
from sanic.request import Request
from sanic.response import HTTPResponse, file, json
from systembridgeshared.const import (
    QUERY_API_KEY,
    QUERY_API_PORT,
    QUERY_AUTOPLAY,
    QUERY_BASE,
    QUERY_FILENAME,
    QUERY_PATH,
    QUERY_URL,
    QUERY_VOLUME,
    SECRET_API_KEY,
    SETTING_ADDITIONAL_MEDIA_DIRECTORIES,
    SETTING_PORT_API,
)
from systembridgeshared.models.media_play import MediaPlay
from systembridgeshared.settings import Settings


def get_directories(settings: Settings) -> list[dict]:
    """Get directories"""
    directories = [
        {
            "key": "documents",
            "path": storagepath.get_documents_dir(),  # type: ignore
        },
        {
            "key": "downloads",
            "path": storagepath.get_downloads_dir(),  # type: ignore
        },
        {
            "key": "home",
            "path": storagepath.get_home_dir(),  # type: ignore
        },
        {
            "key": "music",
            "path": storagepath.get_music_dir(),  # type: ignore
        },
        {
            "key": "pictures",
            "path": storagepath.get_pictures_dir(),  # type: ignore
        },
        {
            "key": "videos",
            "path": storagepath.get_videos_dir(),  # type: ignore
        },
    ]

    additional_directories = settings.get(SETTING_ADDITIONAL_MEDIA_DIRECTORIES)
    if additional_directories is not None and isinstance(additional_directories, list):
        for directory in additional_directories:
            directories.append(
                {
                    "key": directory["name"],
                    "path": directory["value"],
                }
            )

    return directories


def get_files(
    settings: Settings,
    base_path: str,
    path: str,
) -> list[dict]:
    """Get files from path"""
    root_path = None
    for item in get_directories(settings):
        if item["key"] == base_path:
            root_path = item["path"]
            break

    if root_path is None or not os.path.exists(root_path):
        return []

    files_info = []
    for filename in os.listdir(path):
        file_info = get_file(root_path, os.path.join(path, filename))
        if file_info is not None:
            files_info.append(file_info)

    return files_info


def get_file(
    base_path: str,
    filepath: str,
) -> dict | None:
    """Get file from path"""
    try:
        stat = os.stat(filepath)
    except FileNotFoundError:
        return None

    mime_type = None
    if os.path.isfile(filepath):
        mime_type = mimetypes.guess_type(filepath)[0]

    return {
        "name": os.path.basename(filepath),
        "path": filepath.removeprefix(base_path)[1:],
        "fullpath": filepath,
        "size": stat.st_size,
        "last_accessed": stat.st_atime,
        "created": stat.st_ctime,
        "modified": stat.st_mtime,
        "is_directory": os.path.isdir(filepath),
        "is_file": os.path.isfile(filepath),
        "is_link": os.path.islink(filepath),
        "mime_type": mime_type,
    }


async def get_file_data(
    filepath: str,
) -> HTTPResponse:
    """Get file data"""
    return await file(filepath)


async def handler_media_directories(
    _: Request,
    settings: Settings,
) -> HTTPResponse:
    """Handler for media directories"""
    return json(
        {
            "directories": get_directories(settings),
        }
    )


async def handler_media_files(
    request: Request,
    settings: Settings,
) -> HTTPResponse:
    """Handler for media files"""
    if not (query_base := request.args.get(QUERY_BASE)):
        return json(
            {"message": "No base specified"},
            status=400,
        )

    root_path = None
    for item in get_directories(settings):
        if item["key"] == query_base:
            root_path = item["path"]
            break

    if root_path is None or not os.path.exists(root_path):
        return json(
            {"message": "Cannot find base", "base": query_base},
            status=404,
        )

    query_path = request.args.get(QUERY_PATH)
    path = os.path.join(root_path, query_path) if query_path else root_path
    if not os.path.exists(path):
        return json(
            {"message": "Cannot find path", "path": path},
            status=404,
        )
    if not os.path.abspath(path).startswith(os.path.abspath(root_path)):
        return json(
            {
                "message": "Path is not underneath base path",
                "base": root_path,
                "path": path,
            },
            status=400,
        )
    if not os.path.isdir(path):
        return json(
            {"message": "Path is not a directory", "path": path},
            status=400,
        )

    return json(
        {
            "files": get_files(settings, query_base, path),
            "path": path,
        }
    )


async def handler_media_file(
    request: Request,
    settings: Settings,
) -> HTTPResponse:
    """Handler for media file requests"""
    if not (query_base := request.args.get(QUERY_BASE)):
        return json(
            {"message": "No base specified"},
            status=400,
        )

    root_path = None
    for item in get_directories(settings):
        if item["key"] == query_base:
            root_path = item["path"]
            break

    if root_path is None or not os.path.exists(root_path):
        return json(
            {"message": "Cannot find base", "base": query_base},
            status=404,
        )

    if not (query_path := request.args.get(QUERY_PATH)):
        return json(
            {"message": "No path specified"},
            status=400,
        )
    path = os.path.join(root_path, query_path)
    if not os.path.exists(path):
        return json(
            {"message": "Cannot find path", "path": path},
            status=404,
        )
    if not os.path.abspath(path).startswith(os.path.abspath(root_path)):
        return json(
            {
                "message": "Path is not underneath base path",
                "base": root_path,
                "path": path,
            },
            status=400,
        )
    if not os.path.isfile(path):
        return json(
            {"message": "Path is not a file", "path": path},
            status=400,
        )

    return json(get_file(root_path, path))


async def handler_media_file_data(
    request: Request,
    settings: Settings,
) -> HTTPResponse:
    """Handler for media file requests"""
    if not (query_base := request.args.get(QUERY_BASE)):
        return json(
            {"message": "No base specified"},
            status=400,
        )

    root_path = None
    for item in get_directories(settings):
        if item["key"] == query_base:
            root_path = item["path"]
            break

    if root_path is None or not os.path.exists(root_path):
        return json(
            {"message": "Cannot find base", "base": query_base},
            status=404,
        )

    query_path = request.args.get(QUERY_PATH)
    if not (path := os.path.join(root_path, query_path)):
        return json(
            {"message": "Cannot find path", "path": path},
            status=400,
        )
    if not os.path.exists(path):
        return json(
            {"message": "File does not exist", "path": path},
            status=404,
        )
    if not os.path.abspath(path).startswith(os.path.abspath(root_path)):
        return json(
            {
                "message": "Path is not underneath base path",
                "base": root_path,
                "path": path,
            },
            status=400,
        )
    if not os.path.isfile(path):
        return json(
            {"message": "Path is not a file", "path": path},
            status=400,
        )

    return await get_file_data(path)


async def handler_media_file_write(
    request: Request,
    settings: Settings,
) -> HTTPResponse:
    """Handler for media file write requests"""
    if not (query_base := request.args.get(QUERY_BASE)):
        return json(
            {"message": "No base specified"},
            status=400,
        )

    root_path = None
    for item in get_directories(settings):
        if item["key"] == query_base:
            root_path = item["path"]
            break

    if root_path is None or not os.path.exists(root_path):
        return json(
            {"message": "Cannot find base", "base": query_base},
            status=404,
        )

    if not (query_path := request.args.get(QUERY_PATH)):
        return json(
            {"message": "No path specified"},
            status=400,
        )
    if not (query_filename := request.args.get(QUERY_FILENAME)):
        return json(
            {"message": "No filename specified"},
            status=400,
        )
    if not (path := os.path.join(root_path, query_path)):
        return json(
            {"message": "Cannot find path", "path": path},
            status=400,
        )
    if not request.body:
        return json(
            {"message": "No file specified"},
            status=400,
        )

    if not os.path.exists(path):
        os.makedirs(path)
    if not os.path.abspath(path).startswith(os.path.abspath(root_path)):
        return json(
            {
                "message": "Path is not underneath base path",
                "base": root_path,
                "path": path,
            },
            status=400,
        )
    async with aiofiles.open(os.path.join(path, query_filename), "wb") as new_file:
        await new_file.write(request.body)
        await new_file.close()

    return json(
        {
            "message": "File uploaded",
            "path": path,
            "filename": query_filename,
        }
    )


async def handler_media_play(
    request: Request,
    settings: Settings,
    callback: Callable[[str, MediaPlay], None],
) -> HTTPResponse:
    """Handler for media play requests"""
    media_type = request.args.get("type", "video")
    mime_type = None
    path = None
    if "url" not in request.args:
        if not (query_base := request.args.get(QUERY_BASE)):
            return json(
                {"message": "No base specified"},
                status=400,
            )

        root_path = None
        for item in get_directories(settings):
            if item["key"] == query_base:
                root_path = item["path"]
                break

        if root_path is None or not os.path.exists(root_path):
            return json(
                {"message": "Cannot find base", "base": query_base},
                status=404,
            )

        query_path = request.args.get(QUERY_PATH)
        if not (path := os.path.join(root_path, query_path)):
            return json(
                {"message": "Cannot find path", "path": path},
                status=400,
            )
        if not os.path.exists(path):
            return json(
                {"message": "File does not exist", "path": path},
                status=404,
            )
        if not os.path.abspath(path).startswith(os.path.abspath(root_path)):
            return json(
                {
                    "message": "Path is not underneath base path",
                    "base": root_path,
                    "path": path,
                },
                status=400,
            )
        if not os.path.isfile(path):
            return json(
                {"message": "Path is not a file", "path": path},
                status=400,
            )

        url = f"""{request.scheme}://{request.host}/api/media/file/data?{urlencode({
                        QUERY_API_KEY: settings.get_secret(SECRET_API_KEY),
                        QUERY_API_PORT: settings.get(SETTING_PORT_API),
                        QUERY_BASE: query_base,
                        QUERY_PATH: query_path,
                    })}"""

        mime_type, _ = mimetypes.guess_type(path)
        if mime_type is None:
            return json(
                {"message": "Cannot determine mime type", "path": path},
                status=400,
            )

        media_type = (
            "audio"
            if "audio" in mime_type
            else "video"
            if "video" in mime_type
            else None
        )

        if media_type is None:
            return json(
                {
                    "message": "Unsupported file type",
                    "path": path,
                    "mime_type": mime_type,
                },
                status=400,
            )
    else:
        url = request.args.get("url")

        if media_type == "audio":
            path = os.path.join(tempfile.gettempdir(), "tmp.mp3")
            # Download local version to get metadata
            async with ClientSession() as session:
                async with session.get(url) as response:
                    data = await response.read()
                    async with aiofiles.open(
                        path,
                        "wb",
                    ) as new_file:
                        await new_file.write(data)
                        await new_file.close()

    media_play = MediaPlay(
        **{
            QUERY_AUTOPLAY: bool(request.args.get(QUERY_AUTOPLAY, default=False)),
            QUERY_URL: url,
            QUERY_VOLUME: float(request.args.get(QUERY_VOLUME, default=40)),
        }
    )

    api_port = settings.get(SETTING_PORT_API)
    api_key = settings.get_secret(SECRET_API_KEY)

    if media_type == "audio":
        if path is None:
            return json(
                {
                    "message": "Failed to get path for audio file",
                },
                status=400,
            )
        metadata = MutagenFile(path)

        album = metadata.get("album")
        if album is not None and len(album) > 0:
            media_play.album = album[0]
        elif (album := metadata.get("TALB")) is not None:
            media_play.album = album.text[0]

        artist = metadata.get("artist")
        if artist is not None and len(artist) > 0:
            media_play.artist = artist[0]
        elif (artist := metadata.get("TPE1")) is not None:
            media_play.artist = artist.text[0]

        title = metadata.get("title")
        if title is not None and len(title) > 0:
            media_play.title = title[0]
        elif (title := metadata.get("TIT2")) is not None:
            media_play.title = title.text[0]

        # MP3 etc.
        for key in metadata.keys():
            if key.startswith("APIC"):
                if (cover := metadata.get(key)) is not None:
                    cover_filename = _save_cover_from_binary(
                        cover.data,
                        cover.mime,
                        media_play.album,
                    )
                    if cover_filename is not None:
                        media_play.cover = f"""{request.scheme}://{request.host}/api/media/file/data?{urlencode({
                                                QUERY_API_KEY: settings.get_secret(SECRET_API_KEY),
                                                QUERY_API_PORT: settings.get(SETTING_PORT_API),
                                                QUERY_BASE: "pictures",
                                                QUERY_PATH: cover_filename,
                                            })}"""
                        asyncio.create_task(_delete_cover_delayed(cover_filename))
                    break

        # FLAC
        if media_play.cover is None:
            try:
                if (
                    getattr(metadata, "pictures") is not None
                    and len(metadata.pictures) > 0
                ):
                    cover_filename = _save_cover_from_binary(
                        metadata.pictures[0].data,
                        metadata.pictures[0].mime,
                        media_play.album,
                    )
                    if cover_filename is not None:
                        media_play.cover = f"""{request.scheme}://{request.host}/api/media/file/data?{urlencode({
                                                QUERY_API_KEY: settings.get_secret(SECRET_API_KEY),
                                                QUERY_API_PORT: settings.get(SETTING_PORT_API),
                                                QUERY_BASE: "pictures",
                                                QUERY_PATH: cover_filename,
                                            })}"""
                        asyncio.create_task(_delete_cover_delayed(cover_filename))
            except AttributeError:
                pass

    callback(media_type, media_play)

    return json(
        {
            "message": "Opened media player",
            "media_type": media_type,
            "mime_type": mime_type,
            "path": path,
            "player_url": f"""{request.scheme}://{request.host}/app/player/{media_type}.html?{urlencode({
                    QUERY_API_KEY: api_key,
                    QUERY_API_PORT: api_port,
                    **media_play.dict(exclude_none=True),
                })}""",
            **media_play.dict(exclude_none=True),
        }
    )


async def _delete_cover_delayed(
    file_name: str,
    delay: float = 20,
) -> None:
    """Delete cover after delay."""
    await asyncio.sleep(delay)
    file_path = os.path.join(
        storagepath.get_pictures_dir(),  # type: ignore
        file_name,
    )
    if os.path.exists(file_path):
        os.remove(file_path)


def _save_cover_from_binary(
    data: bytes,
    mime_type: str,
    name: str | None,
) -> str:
    """Save cover from binary."""
    cover_extension = mimetypes.guess_extension(mime_type)
    file_name = re.sub(
        "[^-a-zA-Z0-9_.() ]+",
        "",
        f"{name if name is not None else 'unknown'}__{datetime.timestamp(datetime.now())}{cover_extension}",
    )
    file_path = os.path.join(
        storagepath.get_pictures_dir(),  # type: ignore
        file_name,
    )
    with io.open(file_path, "wb") as cover_file:
        cover_file.write(data)
    return file_name



"""System Bridge: Shortcut"""
import platform


def create_shortcuts():
    """Create shortcuts"""
    if "Windows" in platform.system():
        # pylint: disable=import-error, import-outside-toplevel
        from systembridgebackend.shortcut.windows import create_windows_shortcuts

        create_windows_shortcuts()
    elif "Linux" in platform.system():
        # pylint: disable=import-error, import-outside-toplevel
        from systembridgebackend.shortcut.linux import create_linux_shortcuts

        create_linux_shortcuts()



"""System Bridge: Server Handler - Notification"""
from __future__ import annotations

from collections.abc import Callable

from sanic.request import Request
from sanic.response import HTTPResponse, json
from systembridgeshared.models.notification import Notification as NotificationModel
from systembridgeshared.settings import Settings


async def handler_notification(
    request: Request,
    _: Settings,
    callback: Callable[[NotificationModel], None],
) -> HTTPResponse:
    """Send a notification."""
    if request.json is None:
        return json(
            {
                "mesage": "Missing JSON body",
            },
            status=400,
        )

    if "title" not in request.json:
        return json(
            {
                "message": "No title provided",
            },
            status=400,
        )

    try:
        notification = NotificationModel(**request.json)
    except ValueError as error:
        return json(
            {
                "message": str(error),
            },
            status=400,
        )
    if notification.timeout is None:
        notification.timeout = 5

    callback(notification)

    return json(
        {
            "message": "Notification sent",
        }
    )



"""System Bridge: Network"""
from psutil import net_connections, net_if_addrs, net_if_stats, net_io_counters
from psutil._common import sconn, snetio, snicaddr, snicstats
from systembridgeshared.base import Base


class Network(Base):
    """Network"""

    def connections(self) -> list[sconn]:  # pylint: disable=unsubscriptable-object
        """Connections"""
        return net_connections("all")

    def addresses(
        self,
    ) -> dict[str, list[snicaddr]]:  # pylint: disable=unsubscriptable-object
        """Addresses"""
        return net_if_addrs()

    def stats(self) -> dict[str, snicstats]:  # pylint: disable=unsubscriptable-object
        """Stats"""
        return net_if_stats()

    def io_counters(self) -> snetio:  # pylint: disable=unsubscriptable-object
        """IO Counters"""
        return net_io_counters()
